#ifndef __ENET_ENET_H__
#define __ENET_ENET_H__

#ifdef __cplusplus
extern "C"
{
#endif

#include <stdlib.h>

#include "types.h"
#include "protocol.h"
#include "list.h"

#ifdef WIN32
#include "win32.h"
#else
#include "unix.h"
#endif

typedef enum
{
   ENET_SOCKET_TYPE_STREAM   = 1,
   ENET_SOCKET_TYPE_DATAGRAM = 2
} ENetSocketType;

typedef enum
{
   ENET_SOCKET_WAIT_NONE    = 0,
   ENET_SOCKET_WAIT_SEND    = (1 << 0),
   ENET_SOCKET_WAIT_RECEIVE = (1 << 1)
} ENetSocketWait;

enum
{
   ENET_HOST_ANY = 0
};

typedef struct _ENetAddress
{
   uint32 host;
   uint16 port;
} ENetAddress;

typedef enum
{
   ENET_PACKET_FLAG_RELIABLE = (1 << 0)
} ENetPacketFlag;

typedef struct _ENetPacket
{
   size_t               referenceCount;
   uint32               flags;
   uint8 *              data;
   size_t               dataLength;
} ENetPacket;

typedef struct _ENetAcknowledgement
{
   ENetListNode acknowledgementList;
   uint32       sentTime;
   ENetProtocol command;
} ENetAcknowledgement;

typedef struct _ENetOutgoingCommand
{
   ENetListNode outgoingCommandList;
   uint32       reliableSequenceNumber;
   uint32       unreliableSequenceNumber;
   uint32       sentTime;
   uint32       roundTripTimeout;
   uint32       roundTripTimeoutLimit;
   uint32       fragmentOffset;
   uint16       fragmentLength;
   ENetProtocol command;
   ENetPacket * packet;
} ENetOutgoingCommand;

typedef struct _ENetIncomingCommand
{  
   ENetListNode incomingCommandList;
   uint32       reliableSequenceNumber;
   uint32       unreliableSequenceNumber;
   ENetProtocol command;
   uint32       fragmentCount;
   uint32       fragmentsRemaining;
   uint32 *     fragments;
   ENetPacket * packet;
} ENetIncomingCommand;

typedef enum
{
   ENET_PEER_STATE_DISCONNECTED                = 0,
   ENET_PEER_STATE_CONNECTING                  = 1,
   ENET_PEER_STATE_CONNECTED                   = 2,
   ENET_PEER_STATE_DISCONNECTING               = 3,
   ENET_PEER_STATE_ACKNOWLEDGING_DISCONNECTION = 4,
   ENET_PEER_STATE_ZOMBIE                      = 5
} ENetPeerState;

enum
{
   ENET_HOST_RECEIVE_BUFFER_SIZE          = 256 * 1024,
   ENET_HOST_BANDWIDTH_THROTTLE_INTERVAL  = 1000,

   ENET_PEER_DEFAULT_ROUND_TRIP_TIME      = 500,
   ENET_PEER_DEFAULT_PACKET_THROTTLE      = 32,
   ENET_PEER_PACKET_THROTTLE_SCALE        = 32,
   ENET_PEER_PACKET_THROTTLE_COUNTER      = 7, 
   ENET_PEER_PACKET_THROTTLE_ACCELERATION = 7,
   ENET_PEER_PACKET_THROTTLE_DECELERATION = 1,
   ENET_PEER_PACKET_THROTTLE_INTERVAL     = 5000,
   ENET_PEER_PACKET_LOSS_SCALE            = (1 << 16),
   ENET_PEER_PACKET_LOSS_INTERVAL         = 10000,
   ENET_PEER_WINDOW_SIZE_SCALE            = 64 * 1024,
   ENET_PEER_TIMEOUT_LIMIT                = 64,
   ENET_PEER_PING_INTERVAL                = 500
};

typedef struct _ENetChannel
{
   uint32   outgoingReliableSequenceNumber;
   uint32   outgoingUnreliableSequenceNumber;
   uint32   incomingReliableSequenceNumber;
   uint32   incomingUnreliableSequenceNumber;
   ENetList incomingReliableCommands;
   ENetList incomingUnreliableCommands;
} ENetChannel;

typedef struct _ENetPeer
{ 
   struct _ENetHost * host;
   uint16        outgoingPeerID;
   uint16        incomingPeerID;
   uint32        challenge;
   ENetAddress   address;
   void *        data;
   ENetPeerState state;
   ENetChannel * channels;
   size_t        channelCount;
   uint32        incomingBandwidth;
   uint32        outgoingBandwidth;
   uint32        incomingBandwidthThrottleEpoch;
   uint32        outgoingBandwidthThrottleEpoch;
   uint32        incomingDataTotal;
   uint32        outgoingDataTotal;
   uint32        lastSendTime;
   uint32        lastReceiveTime;
   uint32        nextTimeout;
   uint32        packetLossEpoch;
   uint32        packetsSent;
   uint32        packetsLost;
   uint32        packetLoss;
   uint32        packetLossVariance;
   uint32        packetThrottle;
   uint32        packetThrottleLimit;
   uint32        packetThrottleCounter;
   uint32        packetThrottleEpoch;
   uint32        packetThrottleAcceleration;
   uint32        packetThrottleDeceleration;
   uint32        packetThrottleInterval;
   uint32        bestRoundTripTime;
   uint32        roundTripTime;
   uint32        roundTripTimeVariance;
   uint16        packetSize;
   uint32        windowSize;
   uint32        reliableDataInTransit;
   uint32        outgoingReliableSequenceNumber;
   ENetList      acknowledgements;
   ENetList      sentReliableCommands;
   ENetList      sentUnreliableCommands;
   ENetList      outgoingReliableCommands;
   ENetList      outgoingUnreliableCommands;
} ENetPeer;

typedef struct _ENetHost
{
   ENetSocket         socket;
   ENetAddress        address;
   uint32             incomingBandwidth;
   uint32             outgoingBandwidth;
   uint32             bandwidthThrottleEpoch;
   int                recalculateBandwidthLimits;
   ENetPeer *         peers;
   size_t             peerCount;
   ENetPeer *         lastServicedPeer;
   size_t             packetSize;
   ENetProtocol       commands [ENET_PROTOCOL_MAXIMUM_PACKET_COMMANDS];
   size_t             commandCount;
   ENetBuffer         buffers [1 + 2 * ENET_PROTOCOL_MAXIMUM_PACKET_COMMANDS];
   size_t             bufferCount;
   ENetAddress        receivedAddress;
   uint8              receivedData [ENET_PROTOCOL_MAXIMUM_PACKET_SIZE];
   size_t             receivedDataLength;
} ENetHost;

typedef enum
{
   ENET_EVENT_TYPE_NONE       = 0,
   ENET_EVENT_TYPE_CONNECT    = 1,
   ENET_EVENT_TYPE_DISCONNECT = 2,
   ENET_EVENT_TYPE_RECEIVE    = 3
} ENetEventType;

typedef struct _ENetEvent 
{
   ENetEventType        type;
   ENetPeer *           peer;
   uint8                channelID;
   ENetPacket *         packet;
} ENetEvent;

extern int enet_initialize (void);
extern void enet_deinitialize (void);

extern uint32 enet_time_get (void);
extern uint32 enet_time_get_sec (void);
extern void enet_time_set (uint32);

extern ENetSocket enet_socket_create (ENetSocketType, const ENetAddress *);
extern ENetSocket enet_socket_accept (ENetSocket, ENetAddress *);
extern int enet_socket_connect (ENetSocket, const ENetAddress *);
extern int enet_socket_send (ENetSocket, const ENetAddress *, const ENetBuffer *, size_t);
extern int enet_socket_receive (ENetSocket, ENetAddress *, ENetBuffer *, size_t);
extern int enet_socket_wait (ENetSocket, uint32 *, uint32);
extern void enet_socket_destroy (ENetSocket);

extern int enet_address_set_host (ENetAddress *, const char *);
extern int enet_address_get_host (const ENetAddress *, char *, size_t);

extern ENetPacket * enet_packet_create (const void *, size_t, uint32);
extern void enet_packet_destroy (ENetPacket *);
extern int enet_packet_resize  (ENetPacket *, size_t);

extern ENetHost * enet_host_create (const ENetAddress *, size_t, uint32, uint32);
extern void enet_host_destroy (ENetHost *);
extern ENetPeer * enet_host_connect (ENetHost *, const ENetAddress *, size_t);
extern int enet_host_service (ENetHost *, ENetEvent *, uint32);
extern void enet_host_flush (ENetHost *);
extern void enet_host_broadcast (ENetHost *, uint8, ENetPacket *);
extern void enet_host_bandwidth_limit (ENetHost *, uint32, uint32);
extern void enet_host_bandwidth_throttle (ENetHost *);

extern int enet_peer_send (ENetPeer *, uint8, ENetPacket *);
extern ENetPacket * enet_peer_receive (ENetPeer *, uint8);
extern void enet_peer_ping (ENetPeer *);
extern void enet_peer_reset (ENetPeer *);
extern void enet_peer_disconnect (ENetPeer *);
extern void enet_peer_throttle_configure (ENetPeer *, uint32, uint32, uint32);
extern int enet_peer_throttle (ENetPeer *, uint32);
extern ENetOutgoingCommand * enet_peer_queue_outgoing_command (ENetPeer *, const ENetProtocol *, ENetPacket *, uint32, uint16);
extern ENetIncomingCommand * enet_peer_queue_incoming_command (ENetPeer *, const ENetProtocol *, ENetPacket *, uint32);
extern ENetAcknowledgement * enet_peer_queue_acknowledgement (ENetPeer *, const ENetProtocol *, uint32);

#ifdef __cplusplus
}
#endif

#endif /* __ENET_ENET_H__ */

